unit ListModel;

interface

uses
  Classes;

type
  ISubject = interface;

  IObserver = interface
  ['{DF429E76-B8F2-44BE-9395-A40C3ABD0170}']
    procedure UpdateA(Subject: ISubject);
  end;

  ISubject = interface
  ['{721FADCE-372B-4D7C-87F5-18FFF3D79CE8}']
    procedure AttachObserver(Observer: IObserver);
    procedure DetachObserver(Observer: IObserver);
    procedure Notify;
  end;

  IListModel = interface
  ['{E45FA8F8-10B0-4098-9DEC-482E8DB0D928}']
    procedure BeginUpdate;
    procedure EndUpdate;
    function GetCount: Integer;
    function GetItem(Idx: Integer): string;
    procedure Add(Item: string);
    procedure Clear;
    procedure Insert(Item, Before: string);
    procedure Move(Item, Before: string);
    procedure Remove(Item: string);
    property Count: Integer
             read GetCount;
    property Item[Idx: Integer]: string
             read GetItem;
  end;

  TListModel = class(TInterfacedObject, IListModel, ISubject)
  private
    fItems: TStringList;
    fObservers: IInterfaceList;
    fUpdateCount: Integer;
    // IListModel
    procedure BeginUpdate;
    procedure EndUpdate;
    function GetCount: Integer;
    function GetItem(Idx: Integer): string;
    procedure Add(Item: string);
    procedure Clear;
    procedure Insert(Item, Before: string);
    procedure Move(Item, Before: string);
    procedure Remove(Item: string);
    // ISubject
    procedure AttachObserver(Observer: IObserver);
    procedure DetachObserver(Observer: IObserver);
    procedure Notify;
  public
    constructor Create; virtual;
    destructor Destroy; override;
  end;

implementation

{ TListModel }

constructor TListModel.Create;
begin
  inherited Create;
  fItems := TStringList.Create;
  fObservers := TInterfaceList.Create;
end;

destructor TListModel.Destroy;
begin
  fItems.Free;
  inherited Destroy;
end;

function TListModel.GetCount: Integer;
begin
  Result := fItems.Count;
end;

function TListModel.GetItem(Idx: Integer): string;
begin
  Result := fItems[Idx];
end;

procedure TListModel.AttachObserver(Observer: IObserver);
begin
  fObservers.Add(Observer);
  Notify;
end;

procedure TListModel.DetachObserver(Observer: IObserver);
begin
  fObservers.Remove(Observer);
end;

procedure TListModel.Notify;
var
  i: Integer;
begin
  for i := 0 to Pred(fObservers.Count) do
    (fObservers[i] as IObserver).UpdateA(self);
end;

procedure TListModel.Add(Item: string);
begin
  BeginUpdate;
  fItems.Add(Item);
  EndUpdate;
end;

procedure TListModel.Clear;
begin
  BeginUpdate;
  fItems.Clear;
  EndUpdate;
end;

procedure TListModel.Insert(Item, Before: string);
begin
  BeginUpdate;
  fItems.Insert(fItems.IndexOf(Before), Item);
  EndUpdate;
end;

procedure TListModel.Move(Item, Before: string);
var
  IndexOfBefore: Integer;
begin
  BeginUpdate;
  IndexOfBefore := fItems.IndexOf(Before);
  if IndexOfBefore < 0 then
    IndexOfBefore := 0;
  fItems.Delete(fItems.IndexOf(Item));
  fItems.Insert(IndexOfBefore, Item);
  EndUpdate;
end;

procedure TListModel.Remove(Item: string);
begin
  BeginUpdate;
  fItems.Delete(fItems.IndexOf(Item));
  EndUpdate;
end;

procedure TListModel.BeginUpdate;
begin
  Inc(fUpdateCount);
end;

procedure TListModel.EndUpdate;
begin
  Dec(fUpdateCount);
  if fUpdateCount = 0 then
    Notify;
end;

end.
